/*
	Simple Isometric Graphics Library

	Robert L. Watson
	2016-June-26

	http://www.anorthogonaluniverse.com
*/

/*
	Objective:
		Library of functions which allow for manipulation of
		3D primitive solids and render from an "isometric"
		projection (actually 2:1 pixel ratio)
*/

import com.stencyl.graphics.G;
import com.stencyl.models.Scene;
import com.stencyl.models.scene.Layer;
import com.stencyl.Engine;
import com.stencyl.utils.Utils;
import com.stencyl.behavior.Script;
import com.stencyl.behavior.Script.*;

class SimpleIsometric
{
	public static var originX:Float = 0;
    public static var originY:Float = 0;

	public static var offsetX:Float = 0;
	public static var offsetY:Float = 0;
	public static var offsetZ:Float = 0;

	public static var scaleX:Float = 1.0;
	public static var scaleY:Float = 1.0;
	public static var scaleZ:Float = 1.0;

	// Technically cheating; should be "vector", not "vertex"
	// Make sure lightDirection is normalized!
	public static var lightDirection:Vertex = new Vertex( -2 / 2.291287847, -1 / 2.291287847, -0.5 / 2.291287847 );

	public static var maxLight:Float = 0.25;

	public static function setLightDirection( x:Float, y:Float, z:Float )
	{
		lightDirection.x = x;
		lightDirection.y = y;
		lightDirection.z = z;

		lightDirection.doNormalize( );
	}

	public static function setOriginX( x:Float )
	{
		originX = x;
	}

	public static function setOriginY( y:Float )
	{
		originY = y;
	}

	public static function setScaleX( x:Float )
	{
		scaleX = x;
	}

	public static function setScaleY( y:Float )
	{
		scaleY = y;
	}

	public static function setScaleZ( z:Float )
	{
		scaleZ = z;
	}

	public static function setLightScalar( v:Float )
	{
		maxLight = Math.max( 0, Math.min( v, 1 ) );
	}

	public static function doRender( s:Dynamic )
	{
		s.render( );
	}

	public static function doRenderFace( f:Dynamic )
	{
		var vers:Array<Vertex>;

		var intensity:Float;
		var color:Int;

		intensity = SIUtilities.dotProduct( f.normal, SimpleIsometric.lightDirection );
		color = SIUtilities.computeColor( f.color, intensity );
		Engine.engine.g.fillColor = color;

		Engine.engine.g.beginFillPolygon( );

		vers = f.vertices;

		for ( v in vers )
			Engine.engine.g.addPointToPolygon( v.projectToSceneX( ) - getScreenX( ), v.projectToSceneY( ) - getScreenY( ) );

		Engine.engine.g.endDrawingPolygon( );
	}

	/*
		RECTANGLE
    */

	public static function makePyramid2( ):Pyramid
	{
		return new Pyramid( new Vertex( 0, 0, 0 ), 1, 1, 1 );
	}

	/*
		RECTANGLE
    */

	public static function makeRectangularSolid2( ):SolidRectangle
	{
		return new SolidRectangle( new Vertex( 0, 0, 0 ), 1, 1, 1 );
	}


	public static function setWidth( s:Dynamic, val:Float )
	{
		s.setWidth( val );
	}

	public static function setHeight( s:Dynamic, val:Float )
	{
		s.setHeight( val );
	}

	public static function setDepth( s:Dynamic, val:Float )
	{
		s.setDepth( val );
	}

	public static function doTranslate( s:Dynamic, x:Float, y:Float, z:Float )
	{
		s.doTranslate( x, y, z );
	}

	public static function doSetX( s:Dynamic, v:Float )
	{
		s.doTranslate( v - s.vertices[0].x, 0, 0 );
	}

	public static function doSetY( s:Dynamic, v:Float )
	{
		s.doTranslate( 0, v - s.vertices[0].y, 0 );
	}

	public static function doSetZ( s:Dynamic, v:Float )
	{
		s.doTranslate( 0, 0, v - s.vertices[0].z );
	}

	public static function doRotateX2( s:Dynamic, x:Float, y:Float, z:Float, a:Float )
	{
		s.doRotateX2( x, y, z, a );
	}

	public static function doRotateY2( s:Dynamic, x:Float, y:Float, z:Float, a:Float )
	{
		s.doRotateY2( x, y, z, a );
	}

	public static function doRotateZ2( s:Dynamic, x:Float, y:Float, z:Float, a:Float )
	{
		s.doRotateZ2( x, y, z, a );
	}

	public static function doSetColor( s:Dynamic, col:Int )
	{
		s.setColor( col );
	}

	// Return all the renderable faces of the given shape
	public static function getRenderableFaces( solid:Dynamic ):Array<Dynamic>
	{
		var renderableFaces = new Array<Dynamic>( );
		var inlist:Array<Dynamic> = solid.renderableFaces;

		for ( f in inlist )
			renderableFaces.push( f );

		return renderableFaces;
	}

	// Sort the list of renderable faces
	public static function sortRenderableFaces( faces:Dynamic ):Array<Dynamic>
	{
		faces.sort( function( a, b ):Int { if ( a.depthEstimate( ) > b.depthEstimate( ) ) return 1; else return -1; } );

		return faces;
	}

	public static function setOffsetX( val:Float )
	{
		offsetX = val;
	}

	public static function setOffsetY( val:Float )
	{
		offsetY = val;
	}

	public static function setOffsetZ( val:Float )
	{
		offsetZ = val;
	}

	/*
		ADVANCED FUNCTIONS

		These functions are not supplied in block form by default,
		since using them requires custom rendering routines, etc.

		They are mainly supplied for advanced users who would
		like to extend the scope of this extension for their
		own specific purposes.
	*/

	public static function makeVertex( x:Float, y:Float, z:Float ):Vertex
	{
		return new Vertex( x, y, z );
	}

	public static function makeTriangle( v1:Vertex, v2:Vertex, v3:Vertex ):Triangle
	{
		return new Triangle( v1, v2, v3 );
	}

	public static function makeRectangle( v1:Vertex, v2:Vertex, v3:Vertex, v4:Vertex ):Rectangle
	{
		return new Rectangle( v1, v2, v3, v4 );
	}

	public static function setNormal( f:Dynamic )
	{
		f.setNormal( );
	}

	public static function faceProximity( f:Dynamic ):Float
	{
		return f.depthEstimate( );
	}

	// Origin Independent -- compute the proximity baseed on the face's location on-screen
	public static function oiFaceProximity( f:Dynamic ):Float
	{
		var vertices:Array<Vertex> = f.vertices;

		if ( vertices == null )
			return 0;
		else
		{
			var m:Float = 0;
			var i:Float = 0;

			for ( v in vertices )
			{
				m += v.x - offsetX + v.y - offsetY + v.z - offsetZ;
				i ++;
			}

			return m / i;
		}
	}

	public static function makeRectangularSolid( origin:Vertex, width:Float, height:Float, depth:Float ):SolidRectangle
	{
		return new SolidRectangle( origin, width, height, depth );
	}

	public static function makePyramid( origin:Vertex, width:Float, height:Float, depth:Float ):Pyramid
	{
		return new Pyramid( origin, width, height, depth );
	}

	public static function setVertexX( v:Vertex, val:Float )
	{
		v.x = val;
	}

	public static function setVertexY( v:Vertex, val:Float )
	{
		v.y = val;
	}

	public static function setVertexZ( v:Vertex, val:Float )
	{
		v.z = val;
	}

	public static function doRotateX( s:Dynamic, v:Vertex, a:Float )
	{
		s.doRotateX( v, a );
	}

	public static function doRotateY( s:Dynamic, v:Vertex, a:Float )
	{
		s.doRotateY( v, a );
	}

	public static function doRotateZ( s:Dynamic, v:Vertex, a:Float )
	{
		s.doRotateZ( v, a );
	}

	public static function gridToSceneX( x:Float, y:Float, z:Float )
	{
		return ( x - offsetX ) * scaleX - ( y - offsetY ) * scaleY + originX;
	}

	public static function gridToSceneY( x:Float, y:Float, z:Float )
	{
		return ( ( x - offsetX ) * scaleX / 2 ) + ( ( y - offsetY ) * scaleY / 2 ) - ( z - offsetZ ) * scaleZ + originY;
	}

	public static function solidIsOnScreen( s:Dynamic ):Bool
	{
		var faces:Array<Dynamic>;

		faces = s.faces;

		for ( f in faces )
		{	// Return TRUE if any face is on screen
			if ( faceIsOnScreen( f ) )
				return true;
		}

		return false;
	}

	public static function faceIsOnScreen( f:Dynamic ):Bool
	{
		var vers:Array<Vertex>;
		var minsx:Float = Math.POSITIVE_INFINITY;
		var maxsx:Float = Math.NEGATIVE_INFINITY;
		var minsy:Float = Math.POSITIVE_INFINITY;
		var maxsy:Float = Math.NEGATIVE_INFINITY;

		vers = f.vertices;

		for ( v in vers )
		{
			var sx:Float = gridToSceneX( v.x, v.y, v.z );
			var sy:Float = gridToSceneY( v.x, v.y, v.z );

			if ( sx < minsx )	minsx = sx;
			if ( sx > maxsx )	maxsx = sx;
			if ( sy < minsy )	minsy = sy;
			if ( sy > maxsy )	maxsy = sy;
		}

		return !( minsx > getScreenWidth( ) || minsy > getScreenHeight( ) || maxsx < 0 || maxsy < 0 );
	}

	// Compute "center" of face
	public static function faceCenterX( f:Dynamic ):Float
	{
		return f.centerX( );
	}

	public static function faceCenterY( f:Dynamic ):Float
	{
		return f.centerY( );
	}

	public static function faceCenterZ( f:Dynamic ):Float
	{
		return f.centerZ( );
	}
}

class Pyramid
{
	public var faces:Array<Triangle>;
	public var vertices:Array<Vertex>;

	public var bottom:Rectangle;

	public var renderableFaces:Array<Dynamic>;

	public var f3:Rectangle;

	public var color:Int;

	public function new( origin:Vertex, width:Float, height:Float, depth:Float )
	{
		faces = new Array<Triangle>( );
		vertices = new Array<Vertex>( );

		this.color = Utils.getColorRGB( 255, 200, 0 );

		this.vertices.push( origin );
		this.vertices.push( origin.translate( width, 0, 0 ) );
		this.vertices.push( origin.translate( width, height, 0 ) );
		this.vertices.push( origin.translate( 0, height, 0 ) );

		this.vertices.push( origin.translate( width / 2, height / 2, depth ) );

		// Arrange faces in opposite-facing pairs
		faces.push( new Triangle( this.vertices[1], this.vertices[2], this.vertices[4] ) );	// Front Right
		faces.push( new Triangle( this.vertices[3], this.vertices[0], this.vertices[4] ) );	// Back Left

		faces.push( new Triangle( this.vertices[2], this.vertices[3], this.vertices[4] ) );	// Front Left
		faces.push( new Triangle( this.vertices[0], this.vertices[1], this.vertices[4] ) );	// Back Right

		// Create the bottom face
		bottom = new Rectangle( this.vertices[3], this.vertices[2], this.vertices[1], this.vertices[0] );

		this.updateProximity( );
	}

	public function setWidth( val:Float )
	{
		this.vertices[1].x = this.vertices[0].x + val;
		this.vertices[2].x = this.vertices[0].x + val;

		this.vertices[4].x = this.vertices[0].x + val / 2;
	}

	public function setHeight( val:Float )
	{
		this.vertices[2].y = this.vertices[0].y + val;
		this.vertices[3].y = this.vertices[0].y + val;

		this.vertices[4].y = this.vertices[0].y + val / 2;
	}

	public function setDepth( val:Float )
	{
		this.vertices[4].z = this.vertices[0].z + val;
	}

	public function setColor( col:Int )
	{
		this.color = col;

		for ( f in faces )
			f.setColor( col );

		bottom.setColor( col );
	}

	public function doTranslate( x:Float, y:Float, z:Float )
	{
		for ( v in vertices )
			v.doTranslate( x, y, z );
	}

	public function doRotateX( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateX( v2, a );

		this.updateProximity( );
	}

	public function doRotateY( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateY( v2, a );

		this.updateProximity( );
	}

	public function doRotateZ( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ( v2, a );

		this.updateProximity( );
	}

	public function doRotateX2( x:Float, y:Float, z:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateX2( x, y, z, a );

		this.updateProximity( );
	}

	public function doRotateY2( x:Float, y:Float, z:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateY2( x, y, z, a );

		this.updateProximity( );
	}

	public function doRotateZ2( x:Float, y:Float, z:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ2( x, y, z, a );

		this.updateProximity( );
	}

	public function updateProximity( )
	{
		renderableFaces = new Array<Dynamic>( );

		faces.sort( function( a, b ):Int { if ( a.depthEstimate( ) > b.depthEstimate( ) ) return 1; else return -1; } );

		for ( i in 1...faces.length )
		{
			faces[i].setNormal( );

			renderableFaces.push( faces[i] );
		}

		if ( bottom.depthEstimate( ) > vertices[4].x + vertices[4].y + vertices[4].z )
		{
			f3 = bottom;

			f3.setNormal( );

			renderableFaces.push( f3 );
		}
		else
			f3 = null;
	}

	public function render( )
	{
		var intensity:Float;
		var color:Int;

		for ( f in renderableFaces )
		{
			var vers:Array<Vertex>;

			intensity = SIUtilities.dotProduct( f.normal, SimpleIsometric.lightDirection );
			color = SIUtilities.computeColor( this.color, intensity );
			Engine.engine.g.fillColor = color;

			Engine.engine.g.beginFillPolygon( );

			vers = f.vertices;

			for ( v in vers )
				Engine.engine.g.addPointToPolygon( v.projectToSceneX( ) - getScreenX( ), v.projectToSceneY( ) - getScreenY( ) );

			Engine.engine.g.endDrawingPolygon( );
		}
	}
}

// A rectangular solid
class SolidRectangle
{
	public var faces:Array<Rectangle>;
	public var vertices:Array<Vertex>;

	public var renderableFaces:Array<Dynamic>;

	public var f0:Rectangle;
	public var f1:Rectangle;
	public var f2:Rectangle;

	public var color:Int;

	public function new( origin:Vertex, width:Float, height:Float, depth:Float )
	{
		faces = new Array<Rectangle>( );
		vertices = new Array<Vertex>( );

		this.color = Utils.getColorRGB( 255, 200, 0 );

		this.vertices.push( origin );
		this.vertices.push( origin.translate( width, 0, 0 ) );
		this.vertices.push( origin.translate( width, height, 0 ) );
		this.vertices.push( origin.translate( 0, height, 0 ) );

		this.vertices.push( origin.translate( 0, 0, depth ) );
		this.vertices.push( origin.translate( width, 0, depth ) );
		this.vertices.push( origin.translate( width, height, depth ) );
		this.vertices.push( origin.translate( 0, height, depth ) );

		// Arrange faces in opposite-facing pairs, and arrange vertices in:
		//    CLOCKWISE order for originally front-facing faces
		//    COUNTER CLOCKWISE for originally back-facing faces
		faces.push( new Rectangle( this.vertices[2], this.vertices[6], this.vertices[5], this.vertices[1] ) );	// Front Right
		faces.push( new Rectangle( this.vertices[0], this.vertices[4], this.vertices[7], this.vertices[3] ) );	// Back Left
		faces.push( new Rectangle( this.vertices[3], this.vertices[7], this.vertices[6], this.vertices[2] ) );	// Front Left
		faces.push( new Rectangle( this.vertices[1], this.vertices[5], this.vertices[4], this.vertices[0] ) );	// Back Right
		faces.push( new Rectangle( this.vertices[4], this.vertices[5], this.vertices[6], this.vertices[7] ) );	// Top
		faces.push( new Rectangle( this.vertices[3], this.vertices[2], this.vertices[1], this.vertices[0] ) );	// Bottom

		this.updateProximity( );
	}

	public function setWidth( val:Float )
	{
		this.vertices[1].x = this.vertices[0].x + val;
		this.vertices[2].x = this.vertices[0].x + val;
		this.vertices[5].x = this.vertices[0].x + val;
		this.vertices[6].x = this.vertices[0].x + val;
	}

	public function setHeight( val:Float )
	{
		this.vertices[2].y = this.vertices[0].y + val;
		this.vertices[3].y = this.vertices[0].y + val;
		this.vertices[6].y = this.vertices[0].y + val;
		this.vertices[7].y = this.vertices[0].y + val;
	}

	public function setDepth( val:Float )
	{
		this.vertices[4].z = this.vertices[0].z + val;
		this.vertices[5].z = this.vertices[0].z + val;
		this.vertices[6].z = this.vertices[0].z + val;
		this.vertices[7].z = this.vertices[0].z + val;
	}

	public function setColor( col:Int )
	{
		this.color = col;

		for ( f in faces )
			f.setColor( col );
	}

	public function doTranslate( x:Float, y:Float, z:Float )
	{
		for ( v in vertices )
			v.doTranslate( x, y, z );
	}

	public function doRotateX( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateX( v2, a );

		this.updateProximity( );
	}

	public function doRotateY( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateY( v2, a );

		this.updateProximity( );
	}

	public function doRotateZ( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ( v2, a );

		this.updateProximity( );
	}

	public function doRotateX2( x:Float, y:Float, z:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateX2( x, y, z, a );

		this.updateProximity( );
	}

	public function doRotateY2( x:Float, y:Float, z:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateY2( x, y, z, a );

		this.updateProximity( );
	}

	public function doRotateZ2( x:Float, y:Float, z:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ2( x, y, z, a );

		this.updateProximity( );
	}

	// Only three of the six faces will need to be rendered. The other three will be in the back.
	// Opposite-facing pairs are 0-1, 2-3, and 4-5
	// Assign f0, f1, and f2 to the faces to be rendered.
	public function updateProximity( )
	{
		renderableFaces = new Array<Dynamic>( );

		if ( faces[0].depthEstimate( ) > faces[1].depthEstimate( ) )
		{
			f0 = faces[0];
			renderableFaces.push( f0 );
		}
		else
		{
			f0 = faces[1];
			renderableFaces.push( f0 );
		}

		if ( faces[2].depthEstimate( ) > faces[3].depthEstimate( ) )
		{
			f1 = faces[2];
			renderableFaces.push( f1 );
		}
		else
		{
			f1 = faces[3];
			renderableFaces.push( f1 );
		}

		if ( faces[4].depthEstimate( ) > faces[5].depthEstimate( ) )
		{
			f2 = faces[4];
			renderableFaces.push( f2 );
		}
		else
		{
			f2 = faces[5];
			renderableFaces.push( f2 );
		}

		// While we're at it, update the required normals
		f0.setNormal( );
		f1.setNormal( );
		f2.setNormal( );
	}

	public function render( )
	{
		var intensity:Float;
		var color:Int;

		for ( f in renderableFaces )
		{
			var vers:Array<Vertex>;

			intensity = SIUtilities.dotProduct( f.normal, SimpleIsometric.lightDirection );
			color = SIUtilities.computeColor( this.color, intensity );
			Engine.engine.g.fillColor = color;

			Engine.engine.g.beginFillPolygon( );

			vers = f.vertices;

			for ( v in vers )
				Engine.engine.g.addPointToPolygon( v.projectToSceneX( ) - getScreenX( ), v.projectToSceneY( ) - getScreenY( ) );

			Engine.engine.g.endDrawingPolygon( );
		}
	}
}

// A rectangle, for use as a face of a solid
class Rectangle
{
	public var vertices:Array<Vertex>;
	public var normal:Vertex;
	public var color:Int;

	public var originalColor:Int;

	public function new( v1:Vertex, v2:Vertex, v3:Vertex, v4:Vertex )
	{
		vertices = new Array<Vertex>( );

		vertices.push( v1 );
		vertices.push( v2 );
		vertices.push( v3 );
		vertices.push( v4 );

		this.setNormal( );

		this.color = Utils.getColorRGB( 255, 200, 0 );
	}

	public function centerX( ):Float
	{
		return ( vertices[0].x + vertices[2].x ) / 2;
	}

	public function centerY( ):Float
	{
		return ( vertices[0].y + vertices[2].y ) / 2;
	}

	public function centerZ( ):Float
	{
		return ( vertices[0].z + vertices[2].z ) / 2;
	}

	public function setColor( col:Int )
	{
		this.color = col;
	}

	public function setOriginalColor( col:Int )
	{
		this.originalColor = col;
	}

	public function setNormal( )
	{
		this.normal = SIUtilities.crossProduct2( vertices[1].x - vertices[0].x, vertices[1].y - vertices[0].y, vertices[1].z - vertices[0].z,
                                                 vertices[2].x - vertices[1].x, vertices[2].y - vertices[1].y, vertices[2].z - vertices[1].z );

		this.normal.doNormalize( );
	}

	// Returns the depth of the center
	public function depthEstimate( ):Float
	{
		var m:Float = 0;

		for ( v in vertices )
			m += v.x + v.y + v.z;

		return m / 4;
	}

	public function doTranslate( x:Float, y:Float, z:Float )
	{
		for ( v in vertices )
			v.doTranslate( x, y, z );
	}

	public function doRotateX( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateX( v2, a );
	}

	public function doRotateY( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateY( v2, a );
	}

	public function doRotateZ( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ( v2, a );
	}

	public function doRotateX2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateX2( nx, ny, nz, a );
	}

	public function doRotateY2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateY2( nx, ny, nz, a );
	}

	public function doRotateZ2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ2( nx, ny, nz, a );
	}
}

// A triangle, for use as a face of a solid
class Triangle
{
	public var vertices:Array<Vertex>;
	public var normal:Vertex;

	public var color:Int;

	public var originalColor:Int;

	public function new( v1:Vertex, v2:Vertex, v3:Vertex )
	{
		vertices = new Array<Vertex>( );

		vertices.push( v1 );
		vertices.push( v2 );
		vertices.push( v3 );

		this.setNormal( );

		this.color = Utils.getColorRGB( 255, 200, 0 );
	}

	// Not the true center, but easy to compute and functions just as well for our purposes
	public function centerX( ):Float
	{
		return ( vertices[0].x + vertices[1].x + vertices[2].x ) / 3;
	}

	public function centerY( ):Float
	{
		return ( vertices[0].y + vertices[1].y + vertices[2].y ) / 3;
	}

	public function centerZ( ):Float
	{
		return ( vertices[0].z + vertices[1].z + vertices[2].z ) / 3;
	}

	public function setColor( col:Int )
	{
		this.color = col;
	}

	public function setOriginalColor( col:Int )
	{
		this.originalColor = col;
	}

	public function setNormal( )
	{
		this.normal = SIUtilities.crossProduct2( vertices[1].x - vertices[0].x, vertices[1].y - vertices[0].y, vertices[1].z - vertices[0].z,
                                                 vertices[2].x - vertices[1].x, vertices[2].y - vertices[1].y, vertices[2].z - vertices[1].z );

		this.normal.doNormalize( );
	}

	// Returns the depth of the center
	public function depthEstimate( ):Float
	{
		var m:Float = 0;

		for ( v in vertices )
			m += v.x + v.y + v.z;

		return m / 3;
	}

	public function doTranslate( x:Float, y:Float, z:Float )
	{
		for ( v in vertices )
			v.doTranslate( x, y, z );
	}

	public function doRotateX( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateX( v2, a );
	}

	public function doRotateY( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateY( v2, a );
	}

	public function doRotateZ( v2:Vertex, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ( v2, a );
	}

	public function doRotateX2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateX2( nx, ny, nz, a );
	}

	public function doRotateY2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateY2( nx, ny, nz, a );
	}

	public function doRotateZ2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		for ( v in vertices )
			v.doRotateZ2( nx, ny, nz, a );
	}
}

class Vertex
{
	public var x:Float;
	public var y:Float;
	public var z:Float;

	public function new( nx:Float, ny:Float, nz:Float )
	{
		this.x = nx;
		this.y = ny;
		this.z = nz;
	}

	// Project onto scene
	// Returns X coordinate
	public function projectToSceneX( )
	{
		return ( this.x - SimpleIsometric.offsetX ) * SimpleIsometric.scaleX - ( this.y - SimpleIsometric.offsetY ) * SimpleIsometric.scaleY + SimpleIsometric.originX;
	}

	// Returns Y coordinate
	public function projectToSceneY( )
	{
		return ( ( this.x - SimpleIsometric.offsetX ) * SimpleIsometric.scaleX / 2 ) + ( ( this.y - SimpleIsometric.offsetY ) * SimpleIsometric.scaleY / 2 ) - ( this.z - SimpleIsometric.offsetZ ) * SimpleIsometric.scaleZ + SimpleIsometric.originY;
	}

	// If two vertices have the same scene coordinates, the one with the higher proximity will be visible
	public function proximity( ):Float
	{
		return x + y + z;
	}

	/*
		The "do" functions manipulate the actual point
	*/
	// Translate a point
	public function doTranslate( x:Float, y:Float, z:Float )
	{
		this.x += x;
		this.y += y;
		this.z += z;
	}

	// Rotate self around v2 by a radians WRT X axis
	public function doRotateX( v2:Vertex, a:Float )
	{
		this.y -= v2.y;
		this.z -= v2.z;

		var y:Float = this.z * Math.sin( a ) + this.y * Math.cos( a );
		var z:Float = this.z * Math.cos( a ) - this.y * Math.sin( a );

		this.y = y + v2.y;
		this.z = z + v2.z;
	}

	// Rotate self around v2 by a radians WRT Y axis
	public function doRotateY( v2:Vertex, a:Float )
	{
		this.x -= v2.x;
		this.z -= v2.z;

		var x:Float = this.x * Math.cos( a ) - this.z * Math.sin( a );
		var z:Float = this.x * Math.sin( a ) + this.z * Math.cos( a );

		this.x = x + v2.x;
		this.z = z + v2.z;
	}

	// Rotate self around v2 by a radians WRT Z axis
	public function doRotateZ( v2:Vertex, a:Float )
	{
		this.x -= v2.x;
		this.y -= v2.y;

		var x:Float = this.x * Math.cos( a ) - this.y * Math.sin( a );
		var y:Float = this.x * Math.sin( a ) + this.y * Math.cos( a );

		this.x = x + v2.x;
		this.y = y + v2.y;
	}

	public function doRotateX2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		this.y -= ny;
		this.z -= nz;

		var y:Float = this.z * Math.sin( a ) + this.y * Math.cos( a );
		var z:Float = this.z * Math.cos( a ) - this.y * Math.sin( a );

		this.y = y + ny;
		this.z = z + nz;
	}

	public function doRotateY2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		this.x -= nx;
		this.z -= nz;

		var x:Float = this.x * Math.cos( a ) - this.z * Math.sin( a );
		var z:Float = this.x * Math.sin( a ) + this.z * Math.cos( a );

		this.x = x + nx;
		this.z = z + nz;
	}

	public function doRotateZ2( nx:Float, ny:Float, nz:Float, a:Float )
	{
		this.x -= nx;
		this.y -= ny;

		var x:Float = this.x * Math.cos( a ) - this.y * Math.sin( a );
		var y:Float = this.x * Math.sin( a ) + this.y * Math.cos( a );

		this.x = x + nx;
		this.y = y + ny;
	}

	/*
		Return a new point with said transformation
	*/
	// Translate a point
	public function translate( x:Float, y:Float, z:Float ):Vertex
	{
		return new Vertex( this.x + x, this.y + y, this.z + z );
	}

	// Rotate self around v2 by a radians WRT X axis
	public function rotateX( v2:Vertex, a:Float ):Vertex
	{
		this.y -= v2.y;
		this.z -= v2.z;

		var y:Float = this.z * Math.sin( a ) + this.y * Math.cos( a );
		var z:Float = this.z * Math.cos( a ) - this.y * Math.sin( a );

		this.y += v2.y;
		this.z += v2.z;

		return new Vertex( this.x, y, z );
	}

	// Rotate self around v2 by a radians WRT Y axis
	public function rotateY( v2:Vertex, a:Float ):Vertex
	{
		this.x -= v2.x;
		this.z -= v2.z;

		var x:Float = this.x * Math.cos( a ) - this.z * Math.sin( a );
		var z:Float = this.x * Math.sin( a ) + this.z * Math.cos( a );

		this.x += v2.x;
		this.z += v2.z;

		return new Vertex( x, this.y, z );
	}

	// Rotate self around v2 by a radians WRT Z axis
	public function rotateZ( v2:Vertex, a:Float ):Vertex
	{
		this.x -= v2.x;
		this.y -= v2.y;

		var x:Float = this.x * Math.cos( a ) - this.y * Math.sin( a );
		var y:Float = this.x * Math.sin( a ) + this.y * Math.cos( a );

		this.x += v2.x;
		this.y += v2.y;

		return new Vertex( x, y, this.z );
	}

	public function magnitude( )
	{
		return Math.sqrt( this.x * this.x + this.y * this.y + this.z * this.z );
	}

	public function doNormalize( )
	{
		var m:Float = this.magnitude( );

		if ( m == 0 )
			return;

		this.x = this.x / m;
		this.y = this.y / m;
		this.z = this.z / m;

		return;
	}
}

/*
	Some basic utilities (such as math)
*/

class SIUtilities
{
	// Compute the cross product of two position vectors whose heads are vertices v1 and v2 respectively
	public static function crossProduct( v1:Vertex, v2:Vertex ):Vertex
	{
		var x:Float = v1.y * v2.z - v2.y * v1.z;
		var y:Float = v2.x * v1.z - v1.x * v2.z;
		var z:Float = v1.x * v2.y - v2.x * v1.y;

		return new Vertex( x, y, z );
    }

	public static function crossProduct2( v1x:Float, v1y:Float, v1z:Float, v2x:Float, v2y:Float, v2z:Float ):Vertex
	{
		var x:Float = v1y * v2z - v2y * v1z;
		var y:Float = v2x * v1z - v1x * v2z;
		var z:Float = v1x * v2y - v2x * v1y;

		return new Vertex( x, y, z );
    }

	// Compute the length of a position vector whose head is vertex v
	public static function magnitude( v:Vertex ):Float
	{
		return Math.sqrt( v.x * v.x + v.y * v.y + v.z * v.z );
	}

	// Normalize a position vector whose head is vertex v
	public static function normalize( v:Vertex ):Vertex
	{
		var m:Float = magnitude( v );

		if ( m == 0 )
			return new Vertex( 0, 0, 0 );

		var x:Float = v.x / m;
		var y:Float = v.y / m;
		var z:Float = v.z / m;

		return new Vertex( x, y, z );
	}

	// Compute the dot product of two position vectors whose heads are vertices v1 and v2 respectively
	public static function dotProduct( v1:Vertex, v2:Vertex ):Float
	{
		return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
	}

	public static function dotProduct2( v1x:Float, v1y:Float, v1z:Float, v2x:Float, v2y:Float, v2z:Float ):Float
	{
		return v1x * v2x + v1y * v2y + v1z * v2z;
	}

	public static function computeColor( col:Int, intensity:Float ):Int
	{
		var red:Int = Utils.getRed( col );
		var green:Int = Utils.getGreen( col );
		var blue:Int = Utils.getBlue( col );

		red = Std.int( Math.max( 0, Math.min( red * ( 1 + intensity * SimpleIsometric.maxLight ), 255 ) ) );
		green = Std.int( Math.max( 0, Math.min( green * ( 1 + intensity * SimpleIsometric.maxLight ), 255 ) ) );
		blue = Std.int( Math.max( 0, Math.min( blue * ( 1 + intensity * SimpleIsometric.maxLight ), 255 ) ) );

		return Utils.getColorRGB( red, green, blue );
	}
}

